/* s518 middleware */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <sys/ioctl.h>
#include <math.h>
#include <time.h>

#include "s518ioctl.h"
#include "s518mid.h"
#include "list.h"


#define DEV_NAME_ROOT		"/dev/s518_%d"
#define GCS_RETRIES 3 // gage calibration set (GCS) retries
#define MAX_RESET_RETRIES 3
#define MAX_COMMAND_Q  16
#define MAX_RESPONSE_Q 16
#define NUM_SDCS 44        // Number of supported Logical Sensor Definition Codes.

// Diagnostics sub-opcodes ----------------------------------------------------------------

#define DIAG_GETRAWDATA		1		// Return channel's raw data.
#define DIAG_GETEEPROM      2
#define DIAG_GETPRODUCTID	4		// Return product model number.
#define DIAG_GETVERSION		5		// Return firmware version number.
#define DIAG_HISPEEDMODE	8		// Switch to high-speed mode.
// commands
#define CMD_UNSUPPORTED		-1		// Unsupported command
#define CMD_RDCHAN			 0		// Get data from 1 channel
#define CMD_DFNTYP			 1		// Declare sensor type
#define CMD_DFNLIM			 2		// Set alarm limits
#define CMD_RDALRM			 3		// Get alarm flags for 8 channels
#define CMD_RDTEMP			 4		// Get TB temperature(s)
#define CMD_DFNOPN			 5		// Set open-sensor values
#define CMD_TARE			 6		// Tare gage
#define CMD_FILT50			 7		// Set 50Hz filter mode
#define CMD_READ8			 8		// Get data from 8 channels
#define CMD_CAL				10		// Calibration function
#define CMD_DIAGNOSE		11		// Diagnostic function
#define CMD_RDSLOP			12		// Get gage slope/intercept
#define CMD_WTSLOP			13		// Set gage slope/intercept
#define CMD_GZERO			14		// Set gage zero
#define CMD_GSPAN			15		// Set gage span
#define CMD_STFILT			16		// Set filter factor
#define CMD_SLEEP			17		// Switch to low-power standby
#define CMD_WAKEUP			18		// Wake from low-power standby
#define CMD_WCOEFF			19		// Set custom-R coefficients

static const char Cmd518[] = {
	(const char)0x00,	// CMD_RDCHAN	--	Get data from channel
	(const char)0x10,	// CMD_DFNTYP	--	Declare sensor type
	(const char)0x20,	// CMD_DFNLIM	--	Set alarm limits
	(const char)0x30,	// CMD_RDALRM	--	Get alarm flags for 8 channels
	(const char)0x40,	// CMD_RDTEMP	--	Get TB temperature(s)
	(const char)0x50,	// CMD_DFNOPN	--	Set open-sensor values
	(const char)0x70,	// CMD_TARE		--	Tare gage
	(const char)0x48,	// CMD_FILT50	--	Set 50Hz filter mode
	(const char)0x58,	// CMD_READ8	--	Get data from 8 channels
	(const char)0xFF,	//					 spare
	(const char)0xE0,	// CMD_CAL		--	Calibration function
	(const char)0xF0,	// CMD_DIAGNOSE	--	Diagnostic function
	(const char)0x80,	// CMD_RDSLOP	--	Get gage slope/intercept
	(const char)0x90,	// CMD_WTSLOP	--	Set gage slope/intercept
	(const char)0xB0,	// CMD_GZERO	--	Set gage zero
	(const char)0xD0,	// CMD_GSPAN	--	Set gage span
	(const char)0x60,	// CMD_STFILT	--	Set filter factor
	(const char)0x43,	// CMD_SLEEP	--	Switch to low-power standby
	(const char)0x42,	// CMD_WAKEUP	--	Wake from low-power standby
	(const char)0xC0	// CMD_WCOEFF	--	Set custom-R coefficients
};

// Logical-to-physical SDC mapping table.  "**" = Legacy types.
// PhysSDC
static const char Sdc518[] = {
	(const char)0x00,	//  0 : 10V **
	(const char)0x21,	//  1 : 10V
	(const char)0x00,	//  2 : 5V **
	(const char)0x15,	//  3 : 5V
	(const char)0x0E,	//  4 : 1.65V **
	(const char)0x16,	//  5 : 500mV
	(const char)0x17,	//  6 : 100mV
	(const char)0x0D,	//  7 : 80mV **
	(const char)0x01,	//  8 : E
	(const char)0x02,	//  9 : J **
	(const char)0x1B,	// 10 : J
	(const char)0x03,	// 11 : K **
	(const char)0x1C,	// 12 : K
	(const char)0x04,	// 13 : T **
	(const char)0x1D,	// 14 : T
	(const char)0x05,	// 15 : S **
	(const char)0x1E,	// 16 : S
	(const char)0x06,	// 17 : R **
	(const char)0x1F,	// 18 : R
	(const char)0x07,	// 19 : 385 **
	(const char)0x18,	// 20 : 385 platinum
	(const char)0x08,	// 21 : 392 **
	(const char)0x19,	// 22 : 392 platinum
	(const char)0x0B,	// 23 : 10K **
	(const char)0x1A,	// 24 : 10K
	(const char)0x11,	// 25 : 4-20 mA
	(const char)0x0F,	// 26 : Strain/pressure gage
	(const char)0x0A,	// 27 : 400 ohm range
	(const char)0x14,	// 28 : 4K ohm range
	(const char)0x20,	// 29 : 600K ohm range
	(const char)0x0C,	// 30 : Custom resistive sensor
	(const char)0x13,	// 31 : Disabled: remove from scan list
	(const char)0x22,	// 32 : N
	(const char)0x23,	// 33 : C
	(const char)0x24,	// 34 : B
	(const char)0xFF,	// 35 : 120 nickel RTD (STDbus only)
	(const char)0x28,	// 36 : 200 nickel RTD
	(const char)0x29,	// 37 : 1K nickel RTD
	(const char)0x2A,	// 38 : 385 platinum, HiRes (0.0125C)
	(const char)0x2B,	// 39 : 392 platinum, HiRes (0.0125C)
	(const char)0x25,	// 40 : 10K thermistor on fixed 600K ohm range
	(const char)0x2C,	// 41 : 10 ohm copper RTD
	(const char)0x14,	// 42 : 1K nickel RTD (4Kohm range, ext lin)
	(const char)0x14	// 43 : 1K platinum RTD (4Kohm range, ext lin)
};


// Linearization table support --------------------------------------------

struct linear_lookup {				// LINEARIZATION LOOKUP TABLE
	double	Slope;				//  Scalar to map sensor data into table index.
	double	Offset;				//  Offset to map sensor data into table index.
	double	Dat[512];			//  Table data to convert to engineering units.
};

struct sensor_attr {				// SENSOR ATTRIBUTES
	double	Scalar;				//  Scalar used to convert raw binary data into engineering units.
	struct linear_lookup *LinTable;			//  Ptr to linearization LOOKUP table (if non-zero).
};

// Client-side linearization lookup tables.
#include "Typ1kNi2.h"		//   1K Ohm Nickel RTD
#include "Typ1kPt.h"		//   1K Ohm Platinum RTD

#define typNop	( (struct linear_lookup *) 0 )	// Unimplemented

static const struct sensor_attr SensorAttr[] =
{ //   Data   Ptr to
  //  Scalar  LinTbl
	{ 0.0010, typNop },		//  0: 10V **
	{ 0.0005, typNop },		//  1: 10V
	{ 0.0005, typNop },		//  2: 5V **
	{ 0.0002, typNop },		//  3: 5V
	{ 0.0001, typNop },		//  4: 1.65V **
	{ 0.0200, typNop },		//  5: 500mV
	{ 0.0050, typNop },		//  6: 100mV
	{ 0.0100, typNop },		//  7: 80mV **
	{ 0.1000, typNop },		//  8: E
	{ 0.1100, typNop },		//  9: J **
	{ 0.1000, typNop },		// 10: J
	{ 0.1700, typNop },		// 11: K **
	{ 0.1000, typNop },		// 12: K
	{ 0.1500, typNop },		// 13: T **
	{ 0.1000, typNop },		// 14: T 
	{ 0.6000, typNop },		// 15: S **
	{ 0.1000, typNop },		// 16: S
	{ 0.5000, typNop },		// 17: R **
	{ 0.1000, typNop },		// 18: R
	{ 0.1000, typNop },		// 19: 385 **
	{ 0.0500, typNop },		// 20: 385 platinum
	{ 0.1000, typNop },		// 21: 392 **
	{ 0.0500, typNop },		// 22: 392 platinum
	{ 0.0200, typNop },		// 23: 10K **
	{ 0.0100, typNop },		// 24: 10K
	{ 0.0100, typNop },		// 25: 4-20 mA
	{ 1.0000, typNop },		// 26: Strain/pressure gage
	{ 0.0200, typNop },		// 27: 400 ohm range
	{ 0.1250, typNop },		// 28: 4K ohm range
	{31.0000, typNop },		// 29: 600K ohm range
	{ 1.0000, typNop },		// 30: Custom resistive sensor
	{ 1.0000, typNop },		// 31: Disabled: remove from scan list
	{ 0.1000, typNop },		// 32: N
	{ 0.1000, typNop },		// 33: C
	{ 0.1000, typNop },		// 34: B
	{ 0.0100, typNop },		// 35: 120 nickel RTD
	{ 0.0500, typNop },		// 36: 200 nickel RTD
	{ 0.0100, typNop },		// 37: 1K nickel RTD
	{ 0.0125, typNop },		// 38: 385 platinum, HiRes (0.0125C)
	{ 0.0125, typNop },		// 39: 392 platinum, HiRes (0.0125C)
	{ 0.0100, typNop },		// 40: 10K on fixed 600K ohm range
	{ 0.1000, typNop },		// 41: 10 ohm
	// All of the following sensor types are processed as follows:
	//  1. Sensor type is configured as a "fundamental" type, such as V or Ohms.
	//  2. Linearization is performed by the Smart A/D class, rather than on board.
	{ 0.1250, &Typ1kNi2 },	// 42: 1K nickel RTD [4Kohm range]
	{ 0.1250, &Typ1kPt }	// 43: 1K platinum RTD [4Kohm range]
};



/* channel attributes */
struct channel_spec {				// CHANNEL ATTRIBUTES
	short	LogDefn;			//  Logical sensor definition code.
	short	HiLim;				//  High alarm limit.
	short	LoLim;				//  Low alarm limit.
	int	LoLimEnable;		//  High alarm limit enabled.
	int	HiLimEnable;		//  Low alarm limit enabled.
	int	LoAlarm;			//  High alarm detected.
	int	HiAlarm;			//  Low alarm detected.
	int	FailHigh;			//  Fail high (versus low) on open sensor detected.
	char	Filter;				//  Filter factor.
	short	GagemVperV;			//	Gage mV/V * 10 (STDbus products only).
	short	GageMaxLoad;		//	Data value expected at max gage load (STDbus only).
};

struct board_spec {
    int board_state;
    int is_mode_high_speed;
    int is_mode_reject_50hz;
    int fault_flags;
    int num_chans;
    struct channel_spec ch[NUM_518_CHANNELS];
};

struct board_map {
    int handle;
    int board_idx;
    struct list_head list;
};

LIST_HEAD(boardlist);

//static struct channel_spec ch_attr[MAX_518_BOARDS][NUM_518_CHANNELS];
static struct board_spec board_attr[MAX_518_BOARDS];

static int find_idx_from_handle(int handle);

static int retrieve_channel_properties(int handle, int idx, struct channel_spec *ch)
{
    struct s518_channel_prop p;
    int rc;
    p.idx = idx;
    rc = ioctl(handle, S518_IOC_GET_CHANNEL, &p);
    if (rc != 0)
        return rc;

    if (ch == NULL)
        return -EINVAL;
    ch->LogDefn = p.lsdc;
    ch->HiLim = p.HiLim;
    ch->LoLim = p.LoLim;
    ch->Filter = p.Filter;
    ch->FailHigh = p.FailHigh;
    ch->LoLimEnable = p.LoLimEnable;
    ch->HiLimEnable = p.HiLimEnable;
    ch->LoAlarm = p.LoAlarm;
    ch->HiAlarm = p.HiAlarm;
	ch->FailHigh = p.FailHigh;
	ch->Filter = p.Filter;
	ch->GagemVperV = p.GagemVperV;
	ch->GageMaxLoad = p.GageMaxLoad;
    return 0;
}

static int load_board_properties(int handle, int idx)
{
    struct board_spec *bd;
    int rc;
    int i;
    struct s518_board_prop p;
    if (idx >= MAX_518_BOARDS)
        return -EINVAL;
    bd = &board_attr[idx];

    rc = ioctl(handle, S518_IOC_GET_BOARD, &p);
    if (rc != 0) {
        bd->board_state = 0;
        bd->is_mode_high_speed = 0;
        bd->is_mode_reject_50hz = 0;
        bd->num_chans = NUM_518_CHANNELS;
        bd->fault_flags &= ~(FAULT_DAV | FAULT_CRMT | FAULT_LED);        
    } else {
        bd->board_state = p.board_state;
        bd->is_mode_high_speed = p.is_mode_high_speed;
        bd->is_mode_reject_50hz = p.is_mode_reject_50hz;
        bd->num_chans = p.num_chans;
        bd->fault_flags = p.fault_flags;
    }
    
    for (i = 0; i < bd->num_chans; i++) {
        rc = retrieve_channel_properties(handle, i, &bd->ch[i]);
        if (rc != 0)
            return rc;
    }
    return 0;
}

// (Windows: InByteDirect)
static int read_byte_direct(int handle, char *byte)
{
    int rc;
    struct s518_cmd cmd;
    cmd.val = 0xff;
    rc = ioctl(handle, S518_IOC_READ_BYTE, &cmd);
    if (rc != 0)
        return rc;
    *byte = cmd.val;
    return 0;
}

static int write_byte_direct(int handle, unsigned char val)
{
    struct s518_cmd iocmd;
    int rc;
    iocmd.val = val;
    rc = ioctl(handle, S518_IOC_WRITE_BYTE, &iocmd);    
    return rc;
}



/* Subtract the ‘struct timeval’ values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0. */

int timespec_subtract(struct timespec *result, struct timespec *x, struct timespec *y)
{
  /* Perform the carry for the later subtraction by updating y. */
  if (x->tv_nsec < y->tv_nsec) {
    int nsec = (y->tv_nsec - x->tv_nsec) / 1000000000 + 1;
    y->tv_nsec -= 1000000000 * nsec;
    y->tv_sec += nsec;
  }
  if (x->tv_nsec - y->tv_nsec > 1000000000) {
    int nsec = (x->tv_nsec - y->tv_nsec) / 1000000000;
    y->tv_nsec += 1000000000 * nsec;
    y->tv_sec -= nsec;
  }

  /* Compute the time remaining to wait.
     tv_usec is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_nsec = x->tv_nsec - y->tv_nsec;

  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}

double timespec_2_double(struct timespec *t)
{
    double secs;
    secs = t->tv_nsec / 1000000000.0 + t->tv_sec;
    return secs;
}

int timespec_2_ms(struct timespec *t)
{
    int ms;
    ms = t->tv_nsec / 1000000 + t->tv_sec * 1000;
    return ms;
}


// returns 0 on success, error otherwise
int handshakewait(int handle, char mask, char xormask, int mstime)
{
    struct timespec t1;//start time
    struct timespec t2;
    double secs;
    int ms;
    int rc;
    unsigned char lastval;
    struct s518_cmd cmd;
    struct timespec elapsed;
    clock_gettime(CLOCK_MONOTONIC, &t1);

    
    do {
        cmd.val = 0;
        rc = ioctl(handle, S518_IOC_READ_STATUS, &cmd);
        if (rc != 0) {
            return rc;
	}
        if (((cmd.val & mask) ^ xormask ) != 0) {
            return 0;
        }
	lastval = cmd.val;
        clock_gettime(CLOCK_MONOTONIC, &t2);
        timespec_subtract(&elapsed, &t1, &t2);
	
    } while (abs(timespec_2_ms(&elapsed)) < mstime);

    rc = ioctl(handle, S518_IOC_READ_STATUS, &cmd);

    if ((rc == 0) && (((cmd.val & mask) ^ xormask ) != 0)) {
        return 0;
    }
    return -EAGAIN;
}

#define MAXHSTIME 1050 // 1000 milliseconds
int handshake_response(int handle, char *resp, int size)
{
    int i;
    int rc;
    struct s518_cmd cmd;
    for (i = 0; i < size; i++) {
        rc = handshakewait(handle, STAT_DAV, 0, MAXHSTIME);
        if (rc != 0) {
	  printf("STAT_DAV fail: %x\n", rc);
            return rc;
        }
	// short delay
        (void) handshakewait(handle, STAT_DAV, 0, 0);
        rc = ioctl(handle, S518_IOC_READ_BYTE, &cmd);
        if (rc != 0) {
	  printf("read byte fail %x\n", rc);
            return rc;
	}
        resp[i] = cmd.val;
    }
    return 0;
}



int s518_reset_board(int handle)
{
    int idx;
    int rc;
    struct channel_spec *ch;
    struct board_spec *bd;
    struct s518_cmd cmd;
    int status;
    int faults = 0;
    int retry = 0;
    
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    rc = ioctl(handle, S518_IOC_RESET_BOARD, 0);
    if (rc != 0)
      return rc;

    bd = &board_attr[idx];
    bd->fault_flags &= ~(FAULT_DAV | FAULT_CRMT | FAULT_LED);
    return faults;
}


int SendCommand(int handle, char *cmd, int size)
{
    int i;
    struct s518_cmd_full iocmd;
    int rc;
    memcpy(iocmd.cmd, cmd, size);
    iocmd.len = size;
    rc = ioctl(handle, S518_IOC_WRITE_COMMAND, &iocmd);    
    return rc;
}

// 518 functions

int s518_get_ambient_temp(int handle, double *temp)
{
    unsigned char resp[2];
    char cmd[1];
    int rc;
    short tval;
    cmd[0] = Cmd518[CMD_RDTEMP];
    rc = SendCommand(handle, cmd, 1);
    if (rc != 0) {
        return rc;
    }
    // cmd
    rc = handshake_response(handle, resp, 2);
    if (rc != 0)
        return rc;
    tval = resp[0] << 8 | resp[1];
    *temp = tval * 0.1;
    return 0;
}

static double read_real(unsigned char *resp)
{
	// Handshake 4 bytes from Smart A/D board.
	int m0 = (unsigned char)resp[0];
	int m1 = (unsigned char)resp[1];
	int m2 = (unsigned char)resp[2];
	int  xp = (unsigned char)resp[3];
 
	if ( (xp < 28) || ((xp | m0 | m1 | m2) == 0) )
		return 0.0;

	double Sign = (m2 > 127) ? -1.0 : 1.0;

	if (xp > 228)
		return Sign * 1.33E+36;
	else
		return Sign * (double)( ( ( ( (m2 | 0x80L) << 8 ) + m1 ) << 8 ) + m0 ) * pow( 2.0, (double)(xp - 152) );
}

#define CAL_LM335 3

short LimTo32K(double value)
{
    if (value > 32767.0)
        return 32767;
    else if (value < -32768.0)
        return -32768;
    else
        return (short) value;
}

int s518_calibrate(int handle, unsigned int chan, unsigned int range, double val, double *cval)
{
    unsigned char resp[4];
    char cmd[4];
    int rc;
    short calvalue;
    struct s518_channel_prop p;
    struct channel_spec *ch;
    int logsdc;
    int idx;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;
    
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;
    
    ch = &board_attr[idx].ch[chan];

    p.idx = chan;
    rc = ioctl(handle, S518_IOC_GET_CHANNEL, &p);
    if (rc != 0) {
        logsdc = ch->LogDefn;
    } else {
        ch->LogDefn = p.lsdc;
        logsdc = p.lsdc;
    }
    
    calvalue = LimTo32K(val / SensorAttr[logsdc].Scalar);
    cmd[0] = Cmd518[CMD_CAL] | (char) chan;
    cmd[1] = range;
    cmd[2] = (calvalue >> 8) & 0xff;
    cmd[3] = (calvalue & 0xff);
    rc = SendCommand(handle, cmd, 4);
    if (rc != 0) {
        return rc;
    }
    if (range == CAL_LM335) {
        rc = handshake_response(handle, resp, 2);
        if (rc != 0)
            return rc;
        *cval = (double) (resp[0] | (resp[1] << 8));
    } else {
        rc = handshake_response(handle, resp, 4);
        if (rc != 0)
            return rc;
        *cval = read_real(resp);
    }
    return 0;
}


int s518_get_fw_version(int handle, short *ver)
{
    unsigned char resp[2];
    char cmd[3];
    int rc;
    short tval;
    cmd[0] = Cmd518[CMD_DIAGNOSE];
    cmd[1] = DIAG_GETVERSION;
    cmd[2] = 0;    
    rc = SendCommand(handle, cmd, 3);
    if (rc != 0) {
        return rc;
    }
    // cmd
    rc = handshake_response(handle, resp, 2);
    if (rc != 0)
        return rc;
    *ver = resp[0] << 8 | resp[1];
    return 0;
}



int s518_get_eeprom_data(int handle, double *pdata)
{
    unsigned char resp[28];
    char cmd[3];
    int rc;
    short tval;
    int i;
    cmd[0] = Cmd518[CMD_DIAGNOSE];
    cmd[1] = DIAG_GETEEPROM;
    cmd[2] = 0;    
    rc = SendCommand(handle, cmd, 3);
    if (rc != 0) {
        return rc;
    }
    // cmd
    rc = handshake_response(handle, resp, 28);
    if (rc != 0)
        return rc;
    printf("raw response\n");
    for (i = 0; i < 28; i++) {
        printf("%x ", resp[i]);
        if ((i != 0) && (i % 8 == 0))
            printf("\n");
    }
    
    printf("REF_VOLTS_5V: %lf\n", read_real(resp));
    *pdata = read_real(resp);    

    printf("REF_VOLTS_500mV: %lf\n", read_real(resp+4));
    *pdata++ = read_real(resp+4);    

    printf("REF_OHMS_400: %lf\n", read_real(resp+8));
    *pdata++ = read_real(resp+8);    
    
    printf("REF_LM335: %d\n", resp[12] + (resp[13] << 8));//+12
    *pdata++ = resp[12] + (resp[13] << 8);
    
    printf("REF_VOLTS_30MV: %lf\n", read_real(resp+16));//+16
    *pdata++ = read_real(resp+16);
    
    printf("REF_VOLTS_4K: %lf\n", read_real(resp+20));//+20
    *pdata++ = read_real(resp+20);
    
    printf("REF_VOLTS_600K: %lf\n", read_real(resp+24));//+20
    *pdata++ = read_real(resp+24);
    return 0;
}



int s518_get_driver_version(int handle, int *ver)
{
    struct s518_version iover;
    int rc;

    rc = ioctl(handle, S518_IOC_VERSION, &iover);
    if (rc != 0)
        return rc;
    *ver = iover.ver;
    return 0;
}

int s518_set_hi_speed_mode(int handle)
{
    char cmd[3];
    int rc;
    int idx;
    struct board_spec *bd;
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    bd = &board_attr[idx];
    
    cmd[0] = Cmd518[CMD_DIAGNOSE];
    cmd[1] = DIAG_HISPEEDMODE;
    cmd[2] = 0;
    rc = SendCommand(handle, cmd, 3);
    bd->is_mode_high_speed = 1;
    return rc;
}

static void encode_real_cmd(unsigned char *cmd, double n)
{
	int		exponent;
	double	mantissa = frexp( fabs(n), &exponent );

	if (exponent < -127) {
        cmd[0] = 0;
        cmd[1] = 0;
        cmd[2] = 0;
        cmd[3] = 0;
	} else {
		cmd[0] = ( (char)( mantissa * 16777216.0 ) );
		cmd[1] = ( (char)( mantissa *    65536.0 ) );
		cmd[2] = ( (char)( n > 0 ? mantissa * 256.0 - 128.0 : mantissa * 256.0 ) );
		cmd[3] = ( (char)( !n ? 0 : exponent + 128.0 ) );
	}
    return;
}

// verify
int s518_set_coefficients(int handle, unsigned int chan, double c1, double c2, double c3)
{
    unsigned char cmd[MAX_COMMAND_Q];
    int j = 0;
    unsigned char *pcmd;
    int rc;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    // If this Smart A/D model supports custom resistive sensors ...
    if (Cmd518[CMD_WCOEFF] == 0xFF)
        return -EINVAL;
    pcmd = &cmd[0];
    *pcmd++ = Cmd518[CMD_WCOEFF] | chan ;
    encode_real_cmd(pcmd, c1);
    pcmd += 4;
    encode_real_cmd(pcmd, c2);
    pcmd += 4;
    encode_real_cmd(pcmd, c3);
    pcmd += 4;
    rc = SendCommand(handle, cmd, 13);
    return rc;
}

int s518_set_fail_mode(int handle, unsigned int chan, int b_failhi)
{
    int idx;
    char cmd[2];
    int rc;
    struct channel_spec *ch;
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    ch = board_attr[idx].ch;
    // Construct image of OpenType flags byte.
    short ChFirst = chan & 0xF8;
    short ChLast = ChFirst | 7;
    char OpenFlags = 0;
    int c;
    for (c = ChFirst; c <= ChLast; c++ )
        OpenFlags |= ( ( ch[c].FailHigh = (b_failhi != 0) ) << (c & 7) );
    // Set OpenType flags for one bank of 8 channels.
    short g = chan >> 3;
    cmd[0] = Cmd518[CMD_DFNOPN] | g;
    cmd[1] = OpenFlags;
    rc = SendCommand(handle, cmd, 2);
    return rc;
}


int s518_set_filter(int handle, unsigned int chan, char val)
{
    int idx;
    char cmd[2];
    int rc;
    struct channel_spec *ch;
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    ch = &board_attr[idx].ch[chan];
    // cache the filter value
    ch->Filter = val;
    cmd[0] = Cmd518[CMD_STFILT] | (char) chan;
    cmd[1] = val;
    rc = SendCommand(handle, cmd, 2);
    return rc;
}


int s518_set_reject_50hz(int handle)
{
    int idx;
    char cmd[1];
    int rc;
    struct board_spec *bd;

    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    bd = &board_attr[idx];

    cmd[0] = Cmd518[CMD_FILT50];
    rc = SendCommand(handle, cmd, 1);
    if (rc != 0)
        return rc;

    bd->board_state |= STAT_REJECT50HZ;
    bd->is_mode_reject_50hz = 1;
    return 0;
}


static double s518_linearize(short LSDC, short RawDat)
{
    struct linear_lookup *tbl;
	tbl = SensorAttr[LSDC].LinTable;
    // Linearization performed by Smart A/D class:
	if (tbl != NULL) {
		/* Linearize Smart A/D channel data using 512-element lookup table
         * Translate Value so that it falls
         * within range of lookup table Dat[] values.
         */
		register short N = (short)( (double)RawDat * tbl->Slope + tbl->Offset );
		if ( N < 0 )
			N = 0;
		else if ( N > 0x7FBFL )
			N = 0x7FBFL;

		/* Set X such that the translated Value,
         * N, is bounded by Dat[X] and Dat[X+1].
         */
		int X = N >> 6;
		/* Get linearization boundary values from table,
         * interpolate between them, and
		 * return with interpolated engineering units value.
         */
		return tbl->Dat[X] + ( (tbl->Dat[X+1] - tbl->Dat[X]) * (double)(N & 0x3FL) ) / 64.0;
	} else /* Linearization performed by Smart A/D board:*/
		return (double)RawDat * SensorAttr[LSDC].Scalar;	//   Fetch sensor data & cvt to engr units
}

int s518_get_all_raw(int handle, char *resp, int rsize)
{
    char cmd[1];
    int rc;
    short tval;
    cmd[0] = Cmd518[CMD_READ8];
    rc = SendCommand(handle, cmd, 1);
    if (rc != 0) {
        printf("send command fail 0x%x\n", rc);
        return rc;
    }
    // cmd
    rc = handshake_response(handle, resp, rsize);
    return rc;
}


int s518_get_all_raw_data(int handle, short *pdata)
{
    unsigned char resp[16];
    int rc;
    int i, j;
    
    rc = s518_get_all_raw(handle, resp, 16);
    if (rc != 0)
        return rc;

    for (i = 0, j = 0; i < NUM_518_CHANNELS; i++, j+=2) {
        *pdata++ = ((resp[j] << 8) | resp[j+1]);
    }
    return 0;
}


int s518_get_raw_chan_data(int handle, unsigned int channel, short *data)
{
    unsigned char resp[2];
    char cmd[1];
    int rc;
    short tval;
    cmd[0] = (char) channel;
    rc = SendCommand(handle, cmd, 1);
    if (rc != 0) {
        return rc;
    }
    // cmd
    rc = handshake_response(handle, resp, 2);
    if (rc != 0)
        return rc;
    tval = resp[0] << 8 | resp[1];
    *data = tval;
    return 0;

}

int s518_get_sensor_data(int handle, unsigned int chan, double *data)
{
    int rc;
    short sdata;
    int idx;
    int logsdc;
    struct channel_spec *ch;
    struct s518_channel_prop p;
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;
    rc = s518_get_raw_chan_data(handle, chan, &sdata);
    if (rc != 0)
        return rc;
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;
    ch = &board_attr[idx].ch[chan];
    p.idx = chan;
    rc = ioctl(handle, S518_IOC_GET_CHANNEL, &p);
    if (rc != 0) {
        logsdc = ch->LogDefn;
    } else {
        ch->LogDefn = p.lsdc;
        logsdc = p.lsdc;
    }
    *data = s518_linearize(logsdc, sdata);
    return 0;
}

int s518_get_all_sensors(int handle, double *data)
{
    int rc;
    short sdata[NUM_518_CHANNELS];
    int idx;
    int i;
    int logsdc;
    struct channel_spec *ch;
    struct board_spec *bd;
    struct s518_channel_prop p;
   
    rc = s518_get_all_raw_data(handle, sdata);
    if (rc != 0) {
      printf("rc %d\n", rc);
        return rc;
    }
    idx = find_idx_from_handle(handle);
    if (idx < 0) {
      printf("invalid index\n");
        return -EINVAL;
    }
    for (i = 0; i < NUM_518_CHANNELS; i++) {
        bd = &board_attr[idx];
        ch = &bd->ch[i];
        p.idx = i;
        rc = ioctl(handle, S518_IOC_GET_CHANNEL, &p);
        if (rc != 0) {
	  printf("get channel fail\n");
            logsdc = ch->LogDefn;
        } else {
            ch->LogDefn = p.lsdc;
            logsdc = p.lsdc;
        }
        data[i] = s518_linearize(logsdc, sdata[i]);
    }
    return 0;
}



static int find_idx_from_handle(int handle)
{
    struct list_head *pos, *q;
    struct board_map *tmp;
    list_for_each_safe(pos, q, &boardlist) {
        tmp = list_entry(pos, struct board_map, list);
        if (tmp->handle == handle)
            return tmp->board_idx;
    }
    return -1;
}


int s518_open(int board_idx)
{
    char devname[32];
    int handle;
    int idx;
    struct board_map *map;
    if (board_idx > MAX_518_BOARDS)
        return -EINVAL;
    sprintf(devname, DEV_NAME_ROOT, board_idx);
    handle = open(devname, O_RDWR);
    idx = find_idx_from_handle(handle);
    if (idx == -1) {
         (void) load_board_properties(handle, board_idx);
    }
    map = malloc(sizeof(struct board_map));
    if (map == NULL)
        return -ENOMEM;
    INIT_LIST_HEAD(&map->list);
    map->handle = handle;
    map->board_idx = board_idx;
    list_add(&(map->list), &(boardlist));
    //list_for_each_entry(map, &boardlist, list)
    //printf("test %d %d\n", map->board_idx, map->handle);
    //printf("find idx by handle %d\n", find_idx_from_handle(handle));
    return handle;
}

void s518_close(int handle)
{
    struct list_head *pos, *q;
    struct board_map *tmp;
    list_for_each_safe(pos, q, &boardlist) {
        tmp = list_entry(pos, struct board_map, list);
        list_del(pos);
        free(tmp);
    }
    close(handle);
    return;
}

int s518_get_status_flags(int handle, unsigned int *status_flags)
{
    int rc;
    struct s518_cmd cmd;
    int idx;
    struct board_spec *bd;
    
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;
    bd = &board_attr[idx];
    rc = ioctl(handle, S518_IOC_READ_STATUS, &cmd);
    if (rc != 0)
        return rc;
    *status_flags = cmd.val & 0xf0;
    *status_flags |= bd->board_state;
    return 0;
}

int s518_set_low_pwr_standby(int handle, unsigned int standby)
{
    int idx;
    char cmd[1];
    int rc;
    struct board_spec *bd;

    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    bd = &board_attr[idx];
    // cache the filter value
    bd->board_state = standby ? (bd->board_state | STAT_SLEEPING) : (bd->board_state & ~STAT_SLEEPING);
    cmd[0] = Cmd518[standby ? CMD_SLEEP : CMD_WAKEUP];
    rc = SendCommand(handle, cmd, 1);
    return rc;
}

int s518_set_sensor_type(int handle, unsigned int chan, short logsdc)
{
    int idx;
    char cmd[2];
    int rc;
    struct board_spec *bd;
    struct channel_spec *ch;
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    if (logsdc >= NUM_SDCS)
        return -EINVAL;

    if (Sdc518[logsdc] == 0xff)
        return -EINVAL;
    
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    bd = &board_attr[idx];
    ch = &bd->ch[chan];
    ch->LogDefn = logsdc;
    cmd[0] = (Cmd518[CMD_DFNTYP] | ((unsigned char) chan & 0xff));
    cmd[1] = Sdc518[logsdc];
    
    rc = SendCommand(handle, cmd, 2);
    // sleep for 100 ms. allow time for new sensor type to become active
    // before further communication
    usleep(100*1000);
    return rc;
}

int s518_get_sensor_type(int handle, unsigned int chan, short *logsdc)
{
    unsigned char resp[4];
    char cmd[4];
    int rc;
    short calvalue;
    struct s518_channel_prop p;
    struct channel_spec *ch;
    int idx;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;
    
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;
    
    ch = &board_attr[idx].ch[chan];

    p.idx = chan;
    rc = ioctl(handle, S518_IOC_GET_CHANNEL, &p);
    if (rc != 0) {
        return -1;
    } else {
        ch->LogDefn = p.lsdc;
        *logsdc = p.lsdc;
    }
    return 0;
}

// gage functions
int s518_set_gage_cal(int handle, unsigned int chan, double slope, short offset)
{
    int idx;
    unsigned char cmd[2];
    unsigned char resp[6];
    int rc;
    int i;
    double rslope;
    short roffset;
    
    struct board_spec *bd;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    if (Cmd518[CMD_WTSLOP] == 0xff)
        return -EINVAL;
    
    idx = find_idx_from_handle(handle);
    if (idx < 0)
        return -EINVAL;

    bd = &board_attr[idx];

    for (i = 0; i < GCS_RETRIES; i++) {
        cmd[0] = Cmd518[CMD_WTSLOP] | (unsigned char) (chan & 0xff);
        encode_real_cmd(cmd+1, slope);
        cmd[5] = offset & 0xff;
        cmd[6] = ((offset >> 8) & 0xff);
        rc = SendCommand(handle, cmd, 7);
        if (rc != 0)
            return rc;
        // sleep 22 ms
        usleep(22*1000);
        cmd[0] = Cmd518[CMD_RDSLOP] | (unsigned char) (chan & 0xff);
        rc = SendCommand(handle, cmd, 1);
        if (rc != 0)
            return rc;
        rc = handshake_response(handle, resp, 6);
        if (rc != 0)
            return rc;
        rslope = read_real(resp);
        roffset = resp[4] + (resp[5] << 8);
        if (rslope == slope && roffset == offset) {
            return 0;
        }
    }
    return -EAGAIN;
}

int s518_set_gage_span(int handle, unsigned int chan, unsigned int val)
{
    unsigned char cmd[3];
    int rc;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    if (Cmd518[CMD_GSPAN] == 0xff)
        return -EINVAL;
    
    cmd[0] = Cmd518[CMD_GSPAN] | (unsigned char) (chan & 0x0f);
    cmd[1] = (val >> 8) & 0xff;
    cmd[2] = (val & 0xff);
    
    rc = SendCommand(handle, cmd, 3);
    if (rc != 0)
        return rc;
    
    return 0;
}

int s518_set_gage_tare(int handle, unsigned int chan)
{
    unsigned char cmd[2];
    int rc;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    if (Cmd518[CMD_TARE] == 0xff)
        return -EINVAL;
    
    cmd[0] = Cmd518[CMD_TARE] | chan;
    rc = SendCommand(handle, cmd, 1);
    if (rc != 0)
        return rc;
    
    return 0;
}

int s518_set_gage_zero(int handle, unsigned int chan)
{
    unsigned char cmd[2];
    int rc;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    if (Cmd518[CMD_GZERO] == 0xff)
        return -EINVAL;

    cmd[0] = Cmd518[CMD_GZERO] | chan;
    rc = SendCommand(handle, cmd, 1);
    if (rc != 0)
        return rc;
    
    return 0;
}

int s518_get_gage_cal(int handle, unsigned int chan, double *slope, short *offset)
{
    int idx;
    unsigned char cmd[2];
    unsigned char resp[6];
    int rc;
    
    if (chan >= NUM_518_CHANNELS)
        return -EINVAL;

    if (Cmd518[CMD_GSPAN] == 0xff)
        return -EINVAL;
    
    cmd[0] = Cmd518[CMD_RDSLOP] | chan;
    rc = SendCommand(handle, cmd, 1);
    if (rc != 0)
        return rc;

    *slope = read_real(resp);
    *offset = resp[4] + (resp[5] << 8);
    return 0;
}



// diagnostic functions

double s518_diag_func0(int handle, unsigned int lsdc, unsigned int rawdata)
{
    return s518_linearize(lsdc, rawdata);

}

int s518_diag_func1(int handle, short *pdata)
{
    return s518_get_all_raw_data(handle, pdata);
}

int s518_diag_func2(int handle, unsigned int chan, short *pdata)
{
    return s518_get_raw_chan_data(handle, chan, pdata);
}

int s518_diag_func3(int handle, char *byte)
{
    int rc;
    rc = read_byte_direct(handle, byte);
    return rc;
}

int s518_diag_func4(int handle, char val)
{
    return write_byte_direct(handle, val);
}

double s518_diagnostic(int handle, unsigned int ch, unsigned int id,
                       unsigned int prm, unsigned int rsiz, char *buf)
{
    char resp[MAX_RESPONSE_Q];
    char cmd[3];
    int rc;
    short tval;
    int i;
    
    cmd[0] = Cmd518[CMD_DIAGNOSE] | ch;
    cmd[1] = id;
    cmd[2] = prm;
    rc = SendCommand(handle, cmd, 3);
    if (rc != 0)
        return rc;
    // cmd
    if (rsiz > MAX_RESPONSE_Q)
        return -EINVAL;
    rc = handshake_response(handle, resp, rsiz);
    if (rc != 0)
        return rc;
    for (i = 0; i < rsiz; i++)
        buf[i] = resp[i];
    return 0;
}


    
